
from __future__ import print_function # PY2

import __main__
import argparse
import sys
import traceback
import tokenize
from ctypes import pythonapi, POINTER, c_long, cast
from types import CodeType as Code

from . import console, enable, disable
from .info import PY2


inspect_flag = cast(pythonapi.Py_InspectFlag, POINTER(c_long)).contents

def set_inspect_flag(value):
	inspect_flag.value = int(value)


CODE_FIELDS = ["argcount", "kwonlyargcount", "nlocals", "stacksize", 
		"flags", "code", "consts", "names", "varnames", "filename", 
		"name", 	"firstlineno", "lnotab", "freevars", "cellvars"]
if PY2:
	CODE_FIELDS.remove("kwonlyargcount")

def update_code(codeobj, **kwargs):
	def field_values():
		for field in CODE_FIELDS:
			original_value = getattr(codeobj, "co_{}".format(field))
			value = kwargs.get(field, original_value)
			yield value
	
	return Code(*field_values())

def update_code_recursively(codeobj, **kwargs):
	updated = {}
	
	def update(codeobj, **kwargs):
		result = updated.get(codeobj, None)
		if result is not None:
			return result
		
		if any(isinstance(c, Code) for c in codeobj.co_consts):
			consts = tuple(update(c, **kwargs) if isinstance(c, Code) else c
				for c in codeobj.co_consts)
		else:
			consts = codeobj.co_consts
		
		result = update_code(codeobj, consts=consts, **kwargs)
		updated[codeobj] = result
		return result
	
	return update(codeobj, **kwargs)


def get_code(path):
	if PY2:
		from .tokenize_open import read_source_lines
		source = u"".join(read_source_lines(path))
	else:
		with tokenize.open(path) as f: # opens with detected source encoding
			source = f.read()
	
	try:
		code = compile(source, path, "exec", dont_inherit=True)
	except UnicodeEncodeError:
		code = compile(source, "<encoding error>", "exec", dont_inherit=True)
		if PY2:
			path = path.encode("utf-8")
		code = update_code_recursively(code, filename=path)
			# so code constains correct filename (even if it contains Unicode)
			# and tracebacks show contents of code lines
	
	return code


def print_exception_without_first_line(etype, value, tb, limit=None, file=None, chain=True):
	if file is None:
		file = sys.stderr
	
	lines = iter(traceback.TracebackException(
		type(value), value, tb, limit=limit).format(chain=chain))
	
	next(lines)
	for line in lines:
		print(line, file=file, end="")


def run_script(args):
	sys.argv = [args.script] + args.script_arguments
	path = args.script
	__main__.__file__ = path
	
	try:
		code = get_code(path)
	except Exception as e:
		traceback.print_exception(e.__class__, e, None, file=sys.stderr)
	else:
		try:
			exec(code, __main__.__dict__)
		except BaseException as e:
			if not sys.flags.inspect and isinstance(e, SystemExit):
				raise
				
			elif PY2: # Python 2 produces tracebacks in mixed encoding (!)
				etype, e, tb = sys.exc_info()
				for line in traceback.format_exception(etype, e, tb.tb_next):
					line = line.decode("utf-8", "replace")
					try:
						sys.stderr.write(line)
					except UnicodeEncodeError:
						line = line.encode(sys.stderr.encoding, "backslashreplace")
						sys.stderr.write(line)
					
					sys.stderr.flush() # is this needed?
				
			else: # PY3
				traceback.print_exception(e.__class__, e, e.__traceback__.tb_next, file=sys.stderr)

def run_init(args):
	if args.init == "enable":
		enable()
	elif args.init == "disable":
		disable()
	elif args.init == "module":
		__import__(args.module)
	elif args.init == "none":
		pass
	else:
		raise ValueError("unknown runner init mode {}".format(repr(args.init)))

def run_with_custom_repl(args):
	run_init(args)
	
	if args.script:
		run_script(args)
	
	if sys.flags.interactive or not args.script:
		if sys.flags.interactive and not args.script:
			console.print_banner()
		try:
			console.enable()
		finally:
			set_inspect_flag(0)

def run_with_standard_repl(args):
	run_init(args)
	
	if args.script:
		run_script(args)
	
	if sys.flags.interactive and not args.script:
		console.print_banner()

def run_arguments():
	parser = argparse.ArgumentParser(description="Runs a script after customizable initialization. By default, win_unicode_console is enabled.")
	
	init_group = parser.add_mutually_exclusive_group()
	init_group.add_argument(
		"-e", "--init-enable", dest="init", action="store_const", const="enable", 
		help="enable win_unicode_console on init (default)")
	init_group.add_argument(
		"-d", "--init-disable", dest="init", action="store_const", const="disable", 
		help="disable win_unicode_console on init")
	init_group.add_argument(
		"-m", "--init-module", dest="module", 
		help="import the given module on init")
	init_group.add_argument(
		"-n", "--no-init", dest="init", action="store_const", const="none", 
		help="do nothing special on init")
	parser.set_defaults(init="enable")
	
	repl_group = parser.add_mutually_exclusive_group()
	repl_group.add_argument(
		"-s", "--standard-repl", dest="use_repl", action="store_false", 
		help="use the standard Python REPL (default)")
	repl_group.add_argument(
		"-c", "--custom-repl", dest="use_repl", action="store_true", 
		help="use win_unicode_console.console REPL")
	parser.set_defaults(use_repl=False)
	
	parser.add_argument("script", nargs="?")
	parser.add_argument("script_arguments", nargs=argparse.REMAINDER, metavar="script-arguments")
	
	try:
		args = parser.parse_args(sys.argv[1:])
	except SystemExit:
		set_inspect_flag(0)	# don't go interactive after printing help
		raise
	
	if args.module:
		args.init = "module"
	
	if args.use_repl:
		run_with_custom_repl(args)
	else:
		run_with_standard_repl(args)




import sys
import platform


WINDOWS = platform.system().lower() == "windows"
PY2 = sys.version_info.major < 3

def check_Windows():
	current_platform = platform.system()
	
	if not WINDOWS:
		raise RuntimeError("available only for Windows, not {}.".format(current_platform))

def check_PY2():
	if not PY2:
		raise RuntimeError("needed only in Python 2")



"""Backport of tokenize.open from Python 3.5

This is the exact Python 3.5 with the following differences:
 - detect_encoding_ex is detect_encoding from Python 3.5 returning also a bool whether a cookie was found
 - detect_encoding calls detect_encoding_ex, so that its signature is the same as in Python 3.5
 - function read_source_lines was added
"""

from codecs import lookup, BOM_UTF8
from io import TextIOWrapper, open as _builtin_open
import re

re_ASCII = 256 # not present in Python 2
cookie_re = re.compile(r'^[ \t\f]*#.*?coding[:=][ \t]*([-\w.]+)', re_ASCII)
blank_re = re.compile(br'^[ \t\f]*(?:[#\r\n]|$)', re_ASCII)


def _get_normal_name(orig_enc):
	"""Imitates get_normal_name in tokenizer.c."""
	# Only care about the first 12 characters.
	enc = orig_enc[:12].lower().replace("_", "-")
	if enc == "utf-8" or enc.startswith("utf-8-"):
		return "utf-8"
	if enc in ("latin-1", "iso-8859-1", "iso-latin-1") or \
			enc.startswith(("latin-1-", "iso-8859-1-", "iso-latin-1-")):
		return "iso-8859-1"
	return orig_enc


def detect_encoding(readline):
	"""
	The detect_encoding() function is used to detect the encoding that should
	be used to decode a Python source file.  It requires one argument, readline,
	in the same way as the tokenize() generator.
	
	It will call readline a maximum of twice, and return the encoding used
	(as a string) and a list of any lines (left as bytes) it has read in.
	
	It detects the encoding from the presence of a utf-8 bom or an encoding
	cookie as specified in pep-0263.  If both a bom and a cookie are present,
	but disagree, a SyntaxError will be raised.  If the encoding cookie is an
	invalid charset, raise a SyntaxError.  Note that if a utf-8 bom is found,
	'utf-8-sig' is returned.
	
	If no encoding is specified, then the default of 'utf-8' will be returned.
	"""
	
	return detect_encoding_ex(readline)[:2]


def detect_encoding_ex(readline):
	try:
		filename = readline.__self__.name
	except AttributeError:
		filename = None
	bom_found = False
	encoding = None
	default = 'utf-8'
	def read_or_stop():
		try:
			return readline()
		except StopIteration:
			return b''
	
	def find_cookie(line):
		try:
			# Decode as UTF-8. Either the line is an encoding declaration,
			# in which case it should be pure ASCII, or it must be UTF-8
			# per default encoding.
			line_string = line.decode('utf-8')
		except UnicodeDecodeError:
			msg = "invalid or missing encoding declaration"
			if filename is not None:
				msg = '{} for {!r}'.format(msg, filename)
			raise SyntaxError(msg)
		
		match = cookie_re.match(line_string)
		if not match:
			return None
		encoding = _get_normal_name(match.group(1))
		try:
			codec = lookup(encoding)
		except LookupError:
			# This behaviour mimics the Python interpreter
			if filename is None:
				msg = "unknown encoding: " + encoding
			else:
				msg = "unknown encoding for {!r}: {}".format(filename,
						encoding)
			raise SyntaxError(msg)
		
		if bom_found:
			if encoding != 'utf-8':
				# This behaviour mimics the Python interpreter
				if filename is None:
					msg = 'encoding problem: utf-8'
				else:
					msg = 'encoding problem for {!r}: utf-8'.format(filename)
				raise SyntaxError(msg)
			encoding += '-sig'
		return encoding
	
	first = read_or_stop()
	if first.startswith(BOM_UTF8):
		bom_found = True
		first = first[3:]
		default = 'utf-8-sig'
	if not first:
		return default, [], False
	
	encoding = find_cookie(first)
	if encoding:
		return encoding, [first], True
	if not blank_re.match(first):
		return default, [first], False
	
	second = read_or_stop()
	if not second:
		return default, [first], False
	
	encoding = find_cookie(second)
	if encoding:
		return encoding, [first, second], True
	
	return default, [first, second], False


def open(filename):
	"""Open a file in read only mode using the encoding detected by
	detect_encoding().
	"""
	buffer = _builtin_open(filename, 'rb')
	try:
		encoding, lines = detect_encoding(buffer.readline)
		buffer.seek(0)
		text = TextIOWrapper(buffer, encoding, line_buffering=True)
		text.mode = 'r'
		return text
	except:
		buffer.close()
		raise

def read_source_lines(filename):
	buffer = _builtin_open(filename, 'rb')
	try:
		encoding, lines, cookie_present = detect_encoding_ex(buffer.readline)
		buffer.seek(0)
		text = TextIOWrapper(buffer, encoding, line_buffering=True)
		text.mode = 'r'
	except:
		buffer.close()
		raise
	
	with text:
		if cookie_present:
			for i in lines:
				yield text.readline().replace("coding", "Coding")
				# so compile() won't complain about encoding declatation in a Unicode string
				# see 2.7/Python/ast.c:228
		
		for line in text:
			yield line
